#
# Report xarxa neural: dades sintètiques provinents de funcions matemàtiques
#

library("genann")
library("xtable")
options(xtable.only.contents = TRUE)  # no inclure l'entorn "tabuar" en la sortida
options(xtable.sanitize.colnames.function = identity)  # no convertir strings
options(xtable.include.rownames = TRUE)

# produir sempre el mateix resultat
set.seed(2016)


report_synthmath <- function(orig, train, test, plotfile)
{
    starttime = Sys.time()
    model <- ann(x=train$x, y=train$y, hls=c(16, 16), lact="tanh", lr=0.1, mo=0.1, tmse=0.001, maxep=5000)
    difftime <- Sys.time() - starttime
    
    print(model)
    
    # prediccions sobre conjunt train
    train_fitted <- predict(model, newdata=train$x)  # dades del model ajustat
    train_errors <- train$y - train_fitted  # residus
    train_mse = sum(train_errors^2) / nrow(train_errors)

    # prediccions sobre conjunt test
    test_fitted <- predict(model, newdata=test$x)  # dades del model ajustat
    test_errors <- test$y - test_fitted  # residus
    test_mse = sum(test_errors^2) / nrow(test_errors)

    #print(train_mse)
    #print(test_mse)

    pdf(plotfile, width=7, height=4.6)
    par(mfrow=c(2, 3))
    par(mar=c(3, 3, 2, 1), mgp=c(2, 0.7, 0), bty="n")

    plot(orig$x, orig$y, xlim=c(-1, 1), ylim=c(-1, 1), col=rgb(0,0,0,0.4), pch=16, asp=1,
         main="Dades originals", xlab="x originals", ylab="y originals",
         panel.first=grid(col="grey"))

    plot(train$x, train$y, xlim=c(-1, 1), ylim=c(-1, 1), col=rgb(0,0,0,0.4), pch=16, asp=1,
         main="Conjunt d'entrenament", xlab="x d'entrenament", ylab="y d'entrenament",
         panel.first=grid(col="grey"))

    plot(test$x, test_fitted, xlim=c(-1, 1), ylim=c(-1, 1), col=rgb(0,0,0,0.4), pch=16, asp=1,
         main="Prediccions", xlab="x de prova", ylab="prediccions",
         panel.first=grid(col="grey"))

    plot.ts(model$errors, log="y",
            main="Evolució de l'EQM", xlab="epochs", ylab="EQM")

    hist(test_errors,
            main="Histograma de residus", xlab="error de predicció", ylab="freqüència")

    plot(test$x, test_errors, xlim=c(-1, 1), ylim=c(-1, 1), col=rgb(0,0,0,0.4), pch=16, asp=1,
         main="Residus", xlab="x de prova", ylab="error de predicció",
         panel.first=grid(col="grey"))

    dev.off()
    
    r = list(
        train_mse*100,  # en %
        test_mse*100,   # en %
        ifelse(model$converged, "si", "no"),
        model$num_epochs,
        difftime[[1]]
        )
    
    return(r)
}

r1 <- report_synthmath(sinfunc, sinfunc_train, sinfunc_test, "report/ann_synthmath_sinfunc.pdf")
r2 <- report_synthmath(discfunc, discfunc_train, discfunc_test, "report/ann_synthmath_discfunc.pdf")
r3 <- report_synthmath(noisefunc, noisefunc_train, noisefunc_test, "report/ann_synthmath_noisefunc.pdf")

results <- rbind(
	"sin"         = r1,
	"discontinua" = r2,
	"soroll"      = r3
	)

results <- as.data.frame(results)
colnames(results) <- c("EQM entrenament", "EQM test", "Convergeix", "Epochs", "Temps")

results_xtab <- xtable(results, digits=c(0,3,3,0,0,2))

print(results)
print(results_xtab, file="report/ann_synthmath.tex")



