#
# Avaluació i reports dels models de XN de MNIST
#

library("genann")
library("xtable")
options(xtable.only.contents = TRUE)  # no inclure l'entorn "tabuar" en la sortida
options(xtable.sanitize.colnames.function = identity)  # no convertir strings
#options(xtable.sanitize.rownames.function = identity)  # idem, per files
#options(xtable.sanitize.text.function = identity)
options(xtable.include.rownames = FALSE)

# produir sempre el mateix resultat
set.seed(2016)

mnist_eval <- function(name, model, testdata)
{
    fitted <- predict(model, newdata=testdata[1:784])  # prediccions sobre el conjunt de test
    errors <- testdata[785:794] - fitted               # residus
    mse    <- sum(errors^2) / nrow(errors)             # error quadràtic mig
    
    label_t <- NULL
    label_f <- NULL

    for (i in 1:nrow(testdata))
    {
        label_t <- c(label_t, which.max(testdata[i, 785:794]) - 1)
        label_f <- c(label_f, which.max(fitted[i,]) - 1)
    }
    
    # calcular matriu de confusió
    tab <- table(label_t, label_f)
    accuracy <- sum(diag(tab)) / sum(tab)

    # calcular PPV i sensitivitat
    ppv <- list("PPV")
    sen <- list("Sen")
    for (i in 1:length(diag(tab)))
    {
        d <- tab[i,i]
        rowsum <- sum(tab[i,])
        colsum <- sum(tab[,i])
        ppv <- cbind(ppv, d / colsum)
        sen <- cbind(sen, d / rowsum)
    }

    labels <- paste("L", 0:9, sep="")
    
    # exportar la taula de PPV i sensitivitat
    ppvsentab <- as.data.frame(rbind(ppv, sen))
    colnames(ppvsentab) <- c("", labels)
    print(ppvsentab)
    print(xtable(ppvsentab, digits=c(0, 0, 3,3,3,3,3,3,3,3,3,3)),
          file=sprintf("report/gann_test_mnist_ppvsen_%s.tex", name),
          include.rownames=FALSE)
    
    # exportar la matriu de confusió
    colnames(tab) <- labels
    rownames(tab) <- labels
    print(tab)
    print(xtable(tab, digits=c(0, 0,0,0,0,0,0,0,0,0,0)),
          file=sprintf("report/gann_test_mnist_confus_%s.tex", name),
          include.rownames=TRUE)

    # retornar resultats mse i accuracy
    print(mse)
    print(accuracy)
    
    metricstab <<- rbind(metricstab, list(name, accuracy * 100, mse))
}


mnist_test <- readMNIST("data/mnist_t10k-images-idx3-ubyte.gz", "data/mnist_t10k-labels-idx1-ubyte.gz")
#mnist_train <- readMNIST("data/mnist_train-images-idx3-ubyte.gz", "data/mnist_train-labels-idx1-ubyte.gz")

load("mnist_evol.Rdata")    # carrega mnist_evol
load("mnist_models.Rdata")  # carrega model1, model2

metricstab <- NULL
mnist_eval("org1", model1, mnist_test)
mnist_eval("org2", model2, mnist_test)
#mnist_eval("org3", model3, mnist_test)

print(metricstab)

metricstab <- as.data.frame(metricstab)
colnames(metricstab) <- c("XN", "Precisió \\%", "EQM")
print(xtable(metricstab, digits=c(0, 0, 2,3)), file="report/gann_test_mnist_bests.tex")

# produir el gràfic de l'evolució de l'AG
pdf("report/gann_mnist_evol_tsplot.pdf", width=5, height=4)
par(mar=c(3, 3, 2, 1), mgp=c(2, 0.7, 0), bty="n")
plot(mnist_evol, main="Evolució de l'AG", xlab="generació", ylab="fitness")
dev.off()

# produir el gràfic dels entrenaments de la XN
pdf("report/gann_mnist_trainings.pdf", width=7, height=3)
par(mar=c(3, 3, 2, 1), mgp=c(2, 0.7, 0), bty="n")
par(mfrow=c(1, 2))
plot.ts(model1$errors, log="y", main="Organisme 1", xlab="epochs", ylab="EQM")
plot.ts(model2$errors, log="y", main="Organisme 2", xlab="epochs", ylab="EQM")
#plot.ts(model3$errors, log="y", main="Organisme 3", xlab="epochs", ylab="EQM")
dev.off()


