#
# Script que recrea la demo de la memòria
#

install.packages("../genann", repos=NULL)
library("genann")

set.seed(2016)  # intentar produir el mateix resultat, pot dependre
                # del generador de nombres aleatoris del sistema

# generar un dataset sintètic amb entrada propera a 0 i sortida en [0, 1]
data <- data.frame(x=seq(-2.5, 2.5, by=0.02))
data$y <- exp(-data$x^2) + rnorm(length(data$x), sd=0.03)

# desordenar les files aleatoriament
data <- data[sample(nrow(data)),]

# partir en subconjunts d'entrenament i prova
data_train <- data[1:175,]      # 70% aprenentatge AG+XN
data_test  <- data[176:251,]    # 30% validació del model de XN

# partir subconjunt d'entrenament en 2 per l'AG
data_gt <- data_train[1:87,]    # 50% per entrenament de XN en AG
data_gv <- data_train[88:175,]  # 50% per validació de XN en AG


# executar algorisme genètic
evol <- gann(data_gt$x, data_gt$y, data_gv$x, data_gv$y, lact="sigmoid")

# crear el model predictiu amb la xarxa neural configurada per l'AG
model <- ann(data_train$x, data_train$y, lact="sigmoid",
             tmse=evol$besttmse,
             lr=evol$bestlr,
             mo=evol$bestmo,
             hls=evol$besthls)

fitted <- predict(model, newdata=data_test$x)  # realitzar prediccions amb el conjunt test
errors <- data_test$y - fitted                 # residus
print(sum(errors^2) / nrow(errors))            # error quadràtic mig

# comprovar l'ajust
print(head(data.frame(x=data_test$x, y=data_test$y, y_estimat=fitted)))

# gràfics, sense decoracions, per al tutorial

#plot(evol)                        # evolució de l'AG
#plot(model)                       # visualització de la XN
#plot.ts(model$errors, log="y")    # aprenentage de la XN
#plot(data_train$x, data_train$y)  # dades d'entrenament
#plot(data_test$x, fitted)         # dades ajustades


# generar els gràfics per a la memòria

pdf("report/demo_evol_tsplot.pdf", width=7, height=4)
par(mar=c(3, 3, 2, 1), mgp=c(2, 0.7, 0), bty="n")
plot(evol, main="Evolució de l'AG", xlab="generació", ylab="fitness")
dev.off()

pdf("report/demo_nnplot.pdf", width=7, height=4)
par(mar=c(3, 3, 2, 1), mgp=c(2, 0.7, 0), bty="n")
plot(model, main="Xarxa neural", xlab="capa", ylab="neurona")
dev.off()

pdf("report/demo_dataplots.pdf", width=7, height=2.3)
par(mar=c(3, 3, 2, 1), mgp=c(2, 0.7, 0), bty="n")
par(mfrow=c(1, 3))
plot(data_train$x, data_train$y, xlim=c(-2.5, 2.5), ylim=c(0, 1), col=rgb(0,0,0,0.5), pch=16,
     main="Dades d'entrenament", xlab="x originals", ylab="y originals",
     panel.first=grid(col="grey"))
plot(data_test$x, fitted, xlim=c(-2.5, 2.5), ylim=c(0, 1), col=rgb(0,0,0,0.7), pch=16,
     main="Prediccions", xlab="x de prova", ylab="prediccions",
     panel.first=grid(col="grey"))
plot.ts(model$errors, log="y", main="Entrenament de la XN", xlab="epochs", ylab="EQM")
dev.off()


